//
// Simple wrapper for OpenGL textures with different formats
//

#include	<assert.h>
#include	"GlTexture.h"
#include	"IntTexture.h"
#include	"FloatTexture.h"

GlTexture :: GlTexture ( int theWidth, int theHeight, GLenum theFormat, GLenum theIntFormat )
{
	id            = 0;
	width          = theWidth;
	height         = theHeight;
	minFilter      = GL_NEAREST;
	magFilter      = GL_NEAREST;
	wrap [0]       = GL_REPEAT;
	wrap [1]       = GL_REPEAT;
	wrap [2]       = 0;
	anisotropy     = -1;
	bound          = false;
	dirtyFlags     = 0;	
}

GlTexture :: ~GlTexture ()
{
	if ( bound )
		unbind ();
	
	if ( id != 0 && target != 0 )
		glDeleteTextures ( 1, &id );	
}

bool	GlTexture :: create ()
{
	glGenTextures   ( 1, &id );
	glBindTexture   ( target, id );
	glPixelStorei   ( GL_UNPACK_ALIGNMENT, 1 );                         // set 1-byte alignment
	
	if ( wrap [0] )
		glTexParameteri ( target, GL_TEXTURE_WRAP_S, wrap [0] );
		
	if ( wrap [1] )
		glTexParameteri ( target, GL_TEXTURE_WRAP_T, wrap [1] );
		
	if ( wrap [2] )
		glTexParameteri ( target, GL_TEXTURE_WRAP_R, wrap [2] );
		
	glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, minFilter );
	glTexParameteri ( target, GL_TEXTURE_MAG_FILTER, magFilter );
	
	if ( anisotropy != -1 )
		glTexParameteri ( target, GL_TEXTURE_MAX_ANISOTROPY_EXT, anisotropy );

	return true;
}

bool	GlTexture :: bind ()
{
	if ( id == 0 )
		return false;
	
	assert ( !bound );
	
	bound = true;
	
	glBindTexture ( target, id );
	
	if ( dirtyFlags )
	{
		if ( dirtyFlags & minFilterFlag )
			glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, minFilter );
			
		if ( dirtyFlags & magFilterFlag )
			glTexParameteri ( target, GL_TEXTURE_MAG_FILTER, magFilter );
			
		if ( dirtyFlags & wrapFlag )
		{
			if ( wrap [0] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_S, wrap [0] );
				
			if ( wrap [1] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_T, wrap [1] );
				
			if ( wrap [2] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_R, wrap [2] );
		}
			
		if ( dirtyFlags & anisoFlag )
			glTexParameteri ( target, GL_TEXTURE_MAX_ANISOTROPY_EXT, anisotropy );
			
		dirtyFlags = 0;
	}
	
	return true;
}

bool	GlTexture :: unbind ()
{
	assert ( bound );

	bound = false;
	
	glBindTexture ( target, 0 );
	
	return true;
}

bool	GlTexture :: setMinFilter  ( GLenum filter )
{
	if ( filter == minFilter )
		return true;
		
	minFilter = filter;
	
	if ( id != 0 )					// texture has been created
		if ( bound )
			glTexParameteri ( target, GL_TEXTURE_MIN_FILTER, minFilter );
		else
			dirtyFlags |= minFilterFlag;
			
	return true;
}

bool	GlTexture :: setMagFilter  ( GLenum filter )
{
	if ( filter == magFilter )
		return true;
		
	magFilter = filter;
	
	if ( id != 0 )					// texture has been created
		if ( bound )
			glTexParameteri ( target, GL_TEXTURE_MAG_FILTER, magFilter );
		else
			dirtyFlags |= magFilterFlag;
			
	return true;
}

bool	GlTexture :: setAnisotropy ( int aniso )
{
	if ( aniso == anisotropy )
		return true;
		
	anisotropy = aniso;
	
	if ( id != 0 )					// texture has been created
		if ( bound )
		{
			if ( anisotropy != -1 )
				glTexParameteri ( target, GL_TEXTURE_MAX_ANISOTROPY_EXT, anisotropy );
		}
		else
			dirtyFlags |= anisoFlag;
			
	return true;
}

bool	GlTexture :: setWrap ( int s, int t, int r )
{
	wrap [0] = s;
	wrap [1] = t;
	wrap [2] = r;
	
	if ( id != 0 )					// texture has been created
		if ( bound )
		{
			if ( wrap [0] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_S, wrap [0] );
				
			if ( wrap [1] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_T, wrap [1] );
				
			if ( wrap [2] )
				glTexParameteri ( target, GL_TEXTURE_WRAP_R, wrap [2] );
				
		}
		else
			dirtyFlags |= wrapFlag;
			
	return true;
}

bool	GlTexture :: buildMipmaps ()
{
	if ( id == 0 )
		return false;

	bind                ();
	glGenerateMipmapEXT ( target );
		
	return true;
}

int		GlTexture :: maxSize ()
{
	int	maxSize;
	
	glGetIntegerv ( GL_MAX_TEXTURE_SIZE, &maxSize );
	
	return maxSize;
}

//////////////////////////////////////////////////////////////

IntTexture :: IntTexture ( int theWidth, int theHeight, int theNumComponents, int theBpc, bool theSigned ) : GlTexture ( theWidth, theHeight, 0, 0 )
{
	bpc           = theBpc;
	numComponents = theNumComponents;
	sgn           = theSigned;
	format        = GL_RGBA_INTEGER_EXT;

	if ( numComponents == 4 )
	{
		if ( bpc == 8 )
			intFormat = (sgn ? GL_RGBA8I_EXT : GL_RGBA8UI_EXT);
		else
		if ( bpc == 16 )
			intFormat = (sgn ? GL_RGBA16I_EXT : GL_RGBA16UI_EXT);
		else
		if ( bpc == 32 )
			intFormat = (sgn ? GL_RGBA32I_EXT : GL_RGBA32UI_EXT);
	}
	else
	if ( numComponents == 1 )
	{
		if ( bpc == 8 )
			intFormat = (sgn ? GL_ALPHA8I_EXT : GL_ALPHA8UI_EXT);
		else
		if ( bpc == 16 )
			intFormat = (sgn ? GL_ALPHA16I_EXT : GL_ALPHA16UI_EXT);
		else
		if ( bpc == 32 )
			intFormat = (sgn ? GL_ALPHA32I_EXT : GL_ALPHA32UI_EXT);
	}
	
	setMinFilter ( GL_NEAREST );
	setMagFilter ( GL_NEAREST );
}

FloatTexture :: FloatTexture ( int theWidth, int theHeight, int theNumComponents, int theBpc ) : GlTexture ( theWidth, theHeight, 0, 0 )
{
	format        = GL_RGBA_INTEGER_EXT;
	intFormat     = GL_RGBA8UI_EXT;
	bpc           = theBpc;
	numComponents = theNumComponents;

	if ( numComponents == 4 )
	{
		format = GL_RGBA;
		
		if ( isExtensionSupported ( "GL_ARB_texture_float" ) )
			intFormat = (bpc == 16 ? GL_RGBA16F_ARB : GL_RGBA32F_ARB);
		else
		if ( isExtensionSupported ( "GL_NV_float_buffer" ) )
			intFormat = (bpc == 16 ? GL_FLOAT_RGBA16_NV : GL_FLOAT_RGBA32_NV );
		else
		if ( isExtensionSupported ( "GL_ATI_texture_float" ) )
			intFormat = (bpc == 16 ? GL_RGBA_FLOAT16_ATI : GL_RGBA_FLOAT32_ATI);
	}
	else
	if ( numComponents == 1 )
	{
		format = GL_ALPHA;
		
		if ( isExtensionSupported ( "GL_ARB_texture_float" ) )
			intFormat = (bpc == 16 ? GL_ALPHA16F_ARB : GL_ALPHA32F_ARB);
		else
		if ( isExtensionSupported ( "GL_NV_float_buffer" ) )
			intFormat = (bpc == 16 ? GL_FLOAT_R16_NV : GL_FLOAT_R32_NV );
		else
		if ( isExtensionSupported ( "GL_ATI_texture_float" ) )
			intFormat = (bpc == 16 ? GL_ALPHA_FLOAT16_ATI : GL_ALPHA_FLOAT32_ATI);
	}
}