//
//
// CUDA kernal to convert image to grey-scale , keep intact one specific color hue
//

#define	EPS	0.006f

typedef unsigned char byte;

texture<uchar4, 2, cudaReadModeElementType> inTex;

__device__ float3	rgbToHsv ( const float3 c )
{
	float	mn    = min ( min ( c.x, c.y ), c.z );
	float	mx    = max ( max ( c.x, c.y ), c.z );
	float	delta = mx - mn;
	float	h, s;

	if ( mx > 0.001f )
	{
		s = delta / mx;

		if ( c.x == mx )
			h = ( c.y - c.z ) / delta;
		else
		if ( c.y == mx )
			h = 2.0f + ( c.z - c.x ) / delta;
		else
			h = 4.0f + ( c.x - c.y ) / delta;
	}
	else
	{
		s = 0.0f;
		h = 0.0f;
	}

	return make_float3 ( h / 6.0f, s, mx );
}

__global__ void ppKernel ( uchar4 * out, int w, int h )
{
	int	x = blockIdx.x * blockDim.x + threadIdx.x;
	int	y = blockIdx.y * blockDim.y + threadIdx.y;
	int	i = y * w + x;
	
	uchar4 res         = tex2D ( inTex, x, y );
	float  scale       = 1.0f / 255.0f;
	float3 clr         = make_float3 ( res.x * scale, res.y * scale, res.z * scale );
	float3	hsv        = rgbToHsv ( clr );								// convert to HSV
	float	l          = 0.3f * clr.x + 0.59f * clr.y + 0.11f * clr.z;	// intensity of a color

	if ( abs ( hsv.x - 0.077f ) < EPS ) 		// hit hue with enough precision => keep color
		out [i] = res;
	else 										// use greyscale image
	{
		byte r = (byte)( 255.0f * l);
		out [i] = make_uchar4 ( r, r, r, 255 ); //make_uchar4 ( x, y, 100, 100 );
	}
}

extern "C" void doPostprocess ( cudaArray * inMap, cudaArray * outMap, uchar4 * buf, int pitch, int w, int h )
{
	cudaBindTextureToArray ( inTex, inMap );

	dim3 threads ( 16, 16 );
	dim3 blocks  ( w / threads.x, h / threads.y);
	
	ppKernel <<<blocks, threads>>> ( buf, w, h );
	
	cudaMemcpyToArray     ( outMap, 0, 0, buf, w * h * sizeof(uchar4), cudaMemcpyDeviceToDevice );
    cudaThreadSynchronize ();
	cudaUnbindTexture     ( inTex );
}
